# -*- coding: utf-8 -*-
import json
import logging
import requests

from odoo import models, fields, api
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.tools import float_utils
from odoo.exceptions import UserError
from odoo.addons.saas_worker.const import IAP_URL, WEBSITE, APPS_URL

_logger = logging.getLogger(__name__)


# --------------------------------------
#  Validate the transaction in iap-odoo
# --------------------------------------
class SaleOrderLine(models.Model):
    _inherit = 'sale.order.line'

    apps_tx_ref = fields.Char(string='Transaction Reference')
    apps_download_allowed = fields.Boolean('Apps Download allowed', compute='_compute_apps_download_allowed')

    def _is_not_sellable_line(self):
        # do not allow changing quantity for Apps products
        return self.product_id.apps_reference or super()._is_not_sellable_line()

    @api.depends('order_id.invoice_status', 'order_id.invoice_ids.state', 'order_id.tag_ids', 'price_unit')
    def _compute_apps_download_allowed(self):
        apps_block_download_tag = self.env.ref('openerp_apps.apps_block_download')
        for line in self:
            is_free_module = float_utils.float_compare(line.price_unit, 0.0, precision_digits=2) == 0
            is_invoice_paid = line._is_invoice_paid() \
                and line.invoice_status == 'invoiced' \
                and apps_block_download_tag not in line.order_id.tag_ids
            line.apps_download_allowed = line.product_uom_qty >= 1.0 and (is_free_module or is_invoice_paid)

    def _is_invoice_paid(self):
        self.ensure_one()
        # define the payment_state that we consider as paid
        allowed_states = self.env['account.move']._get_paid_payment_states()
        # extract the payment states of every invoices
        payments_state = set(self.order_id.invoice_ids.mapped('payment_state'))
        # check if the states extract are a subset of the allowed states
        return payments_state.issubset(allowed_states)

    def _iap_validate_transaction(self, raise_error=True):
        # send confirmation to IAP when the SO is confirmed -> get the transaction token stored in apps_tx_ref for tracking purpose.
        values = {}

        for line in self:
            apps_tx_ref = line.apps_tx_ref

            if '/' in apps_tx_ref:
                if raise_error:
                    raise UserError('You cannot validate a sale order line that is already linked to a transaction. ref: %s - sale order line: %s' % (apps_tx_ref, line.id))
            else:
                apps_reference = line.product_id.apps_reference
                qty = int(line.product_uom_qty)
                values[line.id] = [apps_tx_ref, apps_reference, qty]

        url = IAP_URL + '/iap/1/odoo_validate'
        req = requests.post(url, data={'values': json.dumps(values), 'raise_error': raise_error}, timeout=15, auth=SaasAuth())
        if req.raise_for_status():
            _logger.error("IAP: transaction validation failed. Error %s for SO_lines: %r", req.reason, self)
            if raise_error:
                req.raise_for_status()

        tokens = req.json()

        for key, token in tokens.items():
            line = self.env['sale.order.line'].browse(int(key))
            line.apps_tx_ref = '/'.join([line.apps_tx_ref, token])

        _logger.info("IAP: the IAP transaction validation succeded for SO_lines: %r", self)
        return True

    @api.model
    def _cron_iap_validate_transactions(self):

        lines = self.env['sale.order.line'].search([
            ('order_id.state', '=', 'sale'),
            ('product_id.categ_id', '=', self.env.ref('openerp_apps.iap_product_category').id),
            ('apps_tx_ref', '!=', False),
            ('apps_tx_ref', 'not ilike', '/')
        ])
        if lines:
            lines._iap_validate_transaction(raise_error=False)


class SaleOrder(models.Model):
    _inherit = 'sale.order'

    def _iap_validation(self):

        for order in self:
            #IAP sales
            iap_order_lines = order.order_line.filtered(lambda l: l.apps_tx_ref and '/' not in l.apps_tx_ref and l.product_id.categ_id.id == self.env.ref('openerp_apps.iap_product_category').id)

            if iap_order_lines:
                iap_order_lines._iap_validate_transaction(raise_error=False)

        return True

    def _send_apps_mail(self):
        """
            Selects the order lines with Apps products and sends a mail with the corresponding download links.
        """
        apps_categ_id = self.env.ref('openerp_apps.apps_product_category').id

        def filter_order_lines(line):
            return line.product_id.categ_id.id == apps_categ_id and line.apps_download_allowed

        for order in self:
            #APPS sales
            apps_order_lines = order.order_line.filtered(filter_order_lines)
            # Send emails to customers so they can easily download their purchase
            if apps_order_lines:
                mail_template = self.env.ref('openerp_apps.download_links_template')

                name_and_links = []
                for line in apps_order_lines:
                    support_link = '%s/web/login?redirect=%s/apps/support/%s/%s/%s' % (APPS_URL, APPS_URL, order.name, line.product_id.default_code, line.product_id.apps_reference)
                    name_and_links.append((line.product_id.name, line.product_id._temp_dl_link(90),support_link))
                render_ctx = dict(
                    self.env.context,
                    email_to=order.partner_id.email,
                    apps_order_lines=name_and_links,
                    website_url= WEBSITE,
                )
                mail_template.with_context(render_ctx).send_mail(order.id)

        return True


    def action_iap_validate_transactions(self):

        unsync_orders = []
        for order in self:
            if order.state != 'sale':
                unsync_orders.append(order.name)

        if unsync_orders:
            raise UserError('IAP: Trying validate lines on a sale order that is not Confirmed or Done.\n Faulty records ' + ', '.join(name for name in unsync_orders))

        return self.mapped('order_line')._iap_validate_transaction(raise_error=False)


    def action_confirm(self):
        # This overide is meant to:
        #  - send the apps mail with the download links on sale orders that only contain free apps.
        #  - Validate the transaction of free IAP purchases
        res = super(SaleOrder, self).action_confirm()
        free_orders = self.filtered(lambda o: o.amount_total == 0)
        free_orders._send_apps_mail()
        free_orders._iap_validation()
        return res
